/*
   netrik -- The ANTRIK Internet Viewer
   Copyright (C) Olaf D. Buddenhagen AKA antrik, et al (see AUTHORS)
   Published under the GNU GPL; see LICENSE for details.
*/
/*
 * forms.c -- helper functions for HTML form handling
 *
 * (C) 2002 antrik
 *     2002 Patrice Neff
 */
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "forms.h"
#include "layout.h"

static void encode_string(const char *string, char **data, int *pos, int *size);     /* URL-encode string */

/* write the value of a given form element into the string */
void set_form(string, link)
const struct String	*string;
const struct Link	*link;
{
   int	data_div;    /* string div showing the form data */
   char	*write_pos;    /* starting position of form value in string */
   
   /* find data_div */
   for(data_div=1; string->div[data_div-1].end<=link->start; ++data_div) {    /* find first div *after* link start (first link div is form marker!) */
#ifdef DEBUG
      if(data_div >= string->div_count) {
	 fprintf(stderr, "internal error: trying to work behind string end (while processing form)\n");
	 exit(100);
      }
#endif
   }

   write_pos=&string->text[string->div[data_div-1].end];

   switch(link->form) {    /* dispatch on form element type */
      case FORM_TEXT:
      case FORM_PASS:
      case FORM_HIDDEN: {
/* Patrice, antrik --> */
	 const int	size=string->div[data_div].end - string->div[data_div-1].end;    /* size of value part displayed in page */
	 const int	value_len=strlen(link->value);

	 if(link->form!=FORM_PASS) {    /* display value */
	    /* copy string, replace control charcters */
	    char	*src, *dst;
	    for(src=link->value, dst=write_pos; *src && dst<write_pos+size; ++src, ++dst)
	       *dst= isprint(*src) ? *src : ' ';
	 } else    /* hide value */
	    memset(write_pos, '*', size);

	 if(size>value_len)
	    memset(write_pos+value_len, '_', size-value_len);    /* pad to size (also overwrites '\0' stored by strncpy()!) */
/* <-- Patrice, antrik */
	 
	 break;
      }    /* FORM_TEXT */

      case FORM_CHECKBOX:
      case FORM_RADIO:
	 *write_pos=link->enabled?'*':'\xa0';
	 break;

      case FORM_OPTION:
      case FORM_MULTIOPTION:
	 *(write_pos-1)=link->enabled?'+':'-';    /* modify char *before* "data_div" */

      case FORM_SUBMIT:
	 break;    /* nothing to store */

      case FORM_NO:
	 break;    /* shouldn't occur */
   }    /* switch form type */

}

/* update the page (item tree) to the new value of a form element given by its
 * link number */
void update_form(layout, link_num)
const struct Layout_data	*layout;
int				link_num;
{
   const struct Link_ptr	*list_entry=&layout->links->link[link_num];    /* entry in link list pointing to form element */
   const struct Item		*item=list_entry->item;    /* (string) item containing form element */
   const struct String		*string=item->data.string;    /* string containing form element */
   const struct Link		*link=&string->link[list_entry->num];    /* link (form) data */

   set_form(string, link);
}

/* find form item of form to which the given (form) link belongs */
struct Item *get_form_item(layout, link_num)
const struct Layout_data	*layout;
int				link_num;    /* link number of form element */
{
   struct Item	*link_item=layout->links->link[link_num].item;

   struct Item	*form_item;
   for(form_item=link_item; form_item->list_next!=NULL; form_item=form_item->parent)    /* search parent form item (go upward until tree top reached) */
      if(form_item->type==ITEM_FORM)    /* found */
	 return form_item;
   return NULL;    /* nothing found */
}

/* prepares a form handle that can be used then to iterate through all form
 * elements using form_next() */
struct Form_handle form_start(form, filter)
const struct Item	*form;    /* form item of desired form */
int			filter;    /* return only "successfull" forms */
{
   struct Form_handle	handle;

   handle.filter=filter;

   handle.form_item=form;
   for(handle.cur_item=(struct Item *)form; handle.cur_item->first_child!=NULL; handle.cur_item=handle.cur_item->first_child);    /* find very first item inside form */
   handle.cur_link=0;

   return handle;
}

/*
 * Iterates through all links of all sub-items of the form item stored in the
 * handle, and returns the link every time a form data element is found. NULL
 * indicates no more form elements.
 *
 * The handle keeps track of which items/links have already been processed.
 *
 * If "filter" was passed to form_start(), only "successful" elements are
 * returned, i.e. the ones that have to be submitted to the server. (Have name,
 * value, and are enabled.)
 */
struct Link *form_next(handle)
struct Form_handle	*handle;
{
   /* find next form control */
   for(; handle->cur_item!=handle->form_item; handle->cur_item=handle->cur_item->list_next) {    /* until back at form item (scan all sub-items) */
      if(handle->cur_item->type==ITEM_TEXT) {    /* can contain links */
	 for(; handle->cur_link < handle->cur_item->data.string->link_count; ++handle->cur_link) {    /* all remaining links in current item */
	    const struct Link	*link=&handle->cur_item->data.string->link[handle->cur_link];

	    if(link->form) {    /* found a form control */
	       if(handle->filter) {    /* only successful controls to be returned -> need to test */
		  if(link->name==NULL || !*link->name    /* no name */
		     || link->value==NULL
		     || !link->enabled
		  )    /* not successful -> ignore */
		     continue;
	       }

	       ++handle->cur_link;    /* proceed with next link next time we are called */
	       return (struct Link *)link;
	    }
	 }
	 handle->cur_link=0;    /* start with first link in next item */
      }    /* ITEM_TEXT */
   }    /* for all sub-items */
   return NULL;    /* no more form controls */
}


/* store a string, escaping characters according to URL encoding */
static void encode_string(string, data, pos, size)
const char	*string;    /* string to encode */
char		**data;    /* where to store */
int		*pos;    /* write position inside "data" */
int		*size;    /* allocated size of "data" */
{
   unsigned char	*chr;    /* currently processed character */

   for(chr=(char *)string; *chr; ++chr) {
      if(*pos>*size-5) {    /* not enough space left (5 chars may follow before next realloc in worst case: "%HH=&") */
	 *data=realloc(*data, *size+=16);
	 if(*data==NULL) {
	    fprintf(stderr, "memory allocation error while preparing form data (in function encode_string())\n");
	    exit(1);
	 }
      }

      if(isascii(*chr) && isgraph(*chr) && !strchr("+=&%#", *chr))    /* normal char */
	 (*data)[(*pos)++]=*chr;
      else if(*chr==' ')
	 (*data)[(*pos)++]='+';
      else {    /* needs to be escaped */
	 (*data)[(*pos)++]='%';
	 (*data)[(*pos)++]=(*chr>>4)+ ((*chr>>4)>9 ? 'A'-10 : '0');    /* upper hex digit */
	 (*data)[(*pos)++]=(*chr&0xf)+ ((*chr&0xf)>9 ? 'A'-10 : '0');    /* lower hex digit */
      }
   }
}

/*
 * store form data in URL-encoded string
 *
 * scans a form (from the item tree) and appends the name/value pair of each
 * successful form control to a URL-encoded string
 */
char *url_encode(form)
const struct Item	*form;    /* points to the desired form's describing item in structure tree */
{
   struct Form_handle	handle=form_start(form, 1);    /* state var for iterating through form elements (only succesful ones) */
   struct Link		*link;    /* link struct of current form element */

   char			*data;    /* URL-encoded form data */
   int			write_pos;    /* current end of "data" string */
   int			size;    /* current allocated size of "data" */

   data=NULL; write_pos=size=0;
   while((link=form_next(&handle)) != NULL) {    /* for all successful form elements */
      encode_string(link->name, &data, &write_pos, &size);
      data[write_pos++]='=';
      encode_string(link->value, &data, &write_pos, &size);
      data[write_pos++]='&';
   }
   
   if(write_pos)    /* anything stored */
      data[write_pos-1]='\0';    /* null-terminate (overwrites last '&'!) */

   return data;
}

/*
 * store form data in MIME-encoded string
 *
 * scans a form (from the item tree) and stores the name/value pair of each
 * successful form control as part of a multipart MIME encoded HTTP request body
 */
char *mime_encode(form)
const struct Item	*form;    /* points to the desired form's describing item in structure tree */
{
   struct Form_handle	handle=form_start(form, 1);    /* state var for iterating through form elements (only succesful ones) */
   struct Link		*link;    /* link struct of current form element */

   char			*data;    /* MIME-encoded form data */
   int			end;    /* current end of "data" string */

   /* store data */
   for(data=NULL, end=0; (link=form_next(&handle)) != NULL; ) {    /* all successful form elements */
      static const char	format[]="--"MIME_BOUNDRY"\r\nContent-Disposition: form-data; name=\"%s\"\r\n\r\n%s\r\n";
      const char	part_len=strlen(format)+strlen(link->name)+strlen(link->value);    /* a few bytes too much, but who cares... */
	       
      data=realloc(data, end+part_len);
      if(data==NULL) {
	 fprintf(stderr, "memory allocation error while preparing form data (in function mime_encode())\n");
	 exit(1);
      }
      end+=snprintf(&data[end], part_len, format, link->name, link->value);
   }

   /* store footer */
   {
      static const char	footer[]="--"MIME_BOUNDRY"--\r\n";

      data=realloc(data, end+sizeof(footer));
      if(data==NULL) {
	 fprintf(stderr, "memory allocation error while preparing form data (in function mime_encode())\n");
	 exit(1);
      }
      memcpy(&data[end], footer, sizeof(footer));
   }

   return data;
}
